from configparser import ConfigParser
from io import StringIO
import os
from pathlib import Path

from click import Context, get_app_dir

LEGACY_CONFIG_DIR = Path.home() / ".jncep"
# appdata on Windows but will be something relevant to the platform for other OS's
APPDATA_CONFIG_DIR = Path(get_app_dir("jncep", roaming=True))


def config_dir():
    if APPDATA_CONFIG_DIR.exists():
        # default if exists (if LEGACY_CONFIG_DIR is still there => will be ignored)
        return APPDATA_CONFIG_DIR
    elif LEGACY_CONFIG_DIR.exists():
        return LEGACY_CONFIG_DIR
    # none exists => will need to be created
    return APPDATA_CONFIG_DIR


def has_config_dir():
    return config_dir().exists()


CONFIG_FILE_NAME = "config.ini"

DEFAULT_CONFIG_FILEPATH = config_dir() / CONFIG_FILE_NAME

TOP_SECTION = "JNCEP"

# also used in the declarations for env var names to use for options
ENVVAR_PREFIX = "JNCEP_"


# TODO error hierarchy with JNCEPError at the top
class InvalidOptionError(Exception):
    pass


def list_available_config_options():
    # to prevent import loops
    from .jncep import main

    with Context(main) as ctx:
        info = ctx.to_info_dict()
        envvars = {}
        _extract_envvars(info, envvars)
        no_prefix_envvars = sorted(
            [
                (ev[len(ENVVAR_PREFIX) :], help_)
                for ev, help_ in envvars.items()
                if ev.startswith(ENVVAR_PREFIX)
            ],
            key=lambda x: x[0],
        )
        no_prefix_envvars = dict(no_prefix_envvars)
        return no_prefix_envvars


def _extract_envvars(info, acc):
    if "envvar" in info and info["envvar"]:
        acc[info["envvar"]] = info.get("help", "")
    else:
        for v in info.values():
            if isinstance(v, dict):
                _extract_envvars(v, acc)
            elif isinstance(v, list):
                for i in v:
                    if isinstance(i, dict):
                        _extract_envvars(i, acc)


def set_config_option(config, option, value):
    option = _validate_option(option)
    config[TOP_SECTION][option] = value
    # returns options since its case may have been modified
    return option


def unset_config_option(config, option):
    option = _validate_option(option)
    is_deleted = config.remove_option(None, option)
    return option, is_deleted


def _validate_option(option):
    # keys in upper case when writing
    option = option.upper()
    allowed_options = list_available_config_options().keys()
    if option not in allowed_options:
        raise InvalidOptionError(
            f"Option '{option}' is not valid. Valid options are: "
            f"{', '.join(allowed_options)}"
        )
    return option


def apply_options_from_config():
    # sets the envvar corresponding to the options in config file
    # needs to be launched before the main click command is launched
    # click will pick up the envvars, in the same way as if they were set outside
    config_manager = ConfigManager()
    config_options = config_manager.read_config_options()
    for option, value in config_options[TOP_SECTION].items():
        option_envvar = f"{ENVVAR_PREFIX}{option}"
        # priority to envvars set outside the config file
        # in that case, ignore the value in the config file
        if os.environ.get(option_envvar) is not None:
            continue
        os.environ[option_envvar] = value


class ConfigManager:
    def __init__(self, config_file_path=None):
        if not config_file_path:
            self.config_file_path = DEFAULT_CONFIG_FILEPATH
        else:
            if config_file_path is Path:
                self.config_file_path = config_file_path
            else:
                self.config_file_path = Path(config_file_path)

    def read_config_options(self):
        config = JNCEPConfigParser()
        if not self.config_file_path.exists():
            # first run ? just return the parser (which will be empty except for the
            # default section)
            return config

        with open(self.config_file_path, "r", encoding="utf-8") as f:
            # add a section transparently to conform to a .ini file
            # it is also the default section in the parser so will be
            # there even if
            config_string = f"[{TOP_SECTION}]\n" + f.read()
            config.read_string(config_string)
            return config

    def write_config_options(self, config):
        # make sure the folder exists
        self._ensure_config_dirpath_exists()

        buffer = StringIO()
        config.write(buffer)
        buffer.seek(0)
        config_str = buffer.read()
        # remove section
        config_str = config_str.replace(f"[{TOP_SECTION}]", "").lstrip()
        with open(self.config_file_path, "w", encoding="utf-8") as f:
            f.write(config_str)

    def _ensure_config_dirpath_exists(self):
        self.config_file_path.parent.mkdir(parents=True, exist_ok=True)


class JNCEPConfigParser(ConfigParser):
    def __init__(self):
        # TOP_SECTION will be automatically created if new file
        super().__init__(default_section=TOP_SECTION, interpolation=None)
        # keys in upper case when reading (instead of default lower)
        self.optionxform = lambda x: x.upper()



from functools import partial
import logging

import click
import trio

from . import options
from .. import core, track, update, utils
from ..config import ENVVAR_PREFIX
from ..trio_utils import coro
from .base import CatchAllExceptionsCommand
from .track import sync_series

logger = logging.getLogger(__name__)
console = utils.getConsole()


@click.command(
    name="update",
    help="Generate EPUB files for new parts of all tracked series (or specific "
    "series if a URL argument is passed)",
    cls=CatchAllExceptionsCommand,
)
@click.argument("jnc_url", metavar="(JNOVEL_CLUB_URL?)", required=False)
@options.login_option
@options.password_option
@options.output_option
@options.byvolume_option
@options.images_option
@options.raw_content_option
@options.no_replace_chars_option
@options.css_option
@click.option(
    "-s",
    "--sync",
    "is_sync",
    is_flag=True,
    help=(
        "Flag to sync tracked series based on series followed on J-Novel Club and "
        "update the new ones from the beginning of the series"
    ),
)
@click.option(
    "-j",
    "--jnc-managed",
    "is_jnc_managed",
    is_flag=True,
    envvar=f"{ENVVAR_PREFIX}JNC_MANAGED",
    help=(
        "Flag to indicate whether to use the series followed on the J-Novel Club "
        "website as the tracking reference for updating (equivalent to "
        "running 'track sync --delete --beginning' followed by 'update')"
    ),
)
@click.option(
    "-w",
    "--whole",
    "is_whole_volume",
    is_flag=True,
    envvar=f"{ENVVAR_PREFIX}WHOLE",
    help=(
        "Flag to indicate whether the whole volume should be regenerated when a "
        "new part is detected during the update"
    ),
)
@click.option(
    "-f",
    "--whole-final",
    "is_whole_volume_on_final_part",
    is_flag=True,
    default=False,
    envvar=f"{ENVVAR_PREFIX}WHOLE_FINAL",
    help=(
        "Flag to indicate whether an EPUB with a complete volume should also be "
        "generated when the final part of the volume is included in the update"
    ),
)
@click.option(
    "-e",
    "--use-events",
    "is_use_events",
    is_flag=True,
    default=False,
    envvar=f"{ENVVAR_PREFIX}USE_EVENTS",
    help="Flag to use the events feed to check for updates",
)
@click.pass_context
@coro
async def update_tracked(
    ctx,
    jnc_url,
    email,
    password,
    output_dirpath,
    is_by_volume,
    is_extract_images,
    is_extract_content,
    is_not_replace_chars,
    style_css_path,
    is_sync,
    is_whole_volume,
    is_whole_volume_on_final_part,
    is_use_events,
    is_jnc_managed,
):

    async with core.JNCEPSession(email, password) as session:
        if is_jnc_managed:
            # run the equivalent of:
            # track sync --delete --beginning
            # update

            # prune series that are not followed anymore + track from beginning
            console.info("[important]track sync --delete --beginning[/]")
            # run_sync because we are already in a trio context so we wouldn't be able
            # to nest another trio.run (inside the click command) othwerwise
            await trio.to_thread.run_sync(
                partial(
                    ctx.invoke,
                    sync_series,
                    email=email,
                    password=password,
                    is_delete=True,
                    is_beginning=True,
                )
            )

            # after that, let update run normally
            console.info("[important]update[/]")
            # set sync to False because doesn't make sense to sync again
            # TODO log to the user ?
            is_sync = False

        epub_generation_options = core.EpubGenerationOptions(
            output_dirpath,
            is_by_volume,
            is_extract_images,
            is_extract_content,
            is_not_replace_chars,
            style_css_path,
        )

        update_options = update.UpdateOptions(
            is_sync,
            is_whole_volume,
            is_whole_volume_on_final_part,
            is_use_events,
        )

        track_manager = track.TrackConfigManager()
        tracked_series = track_manager.read_tracked_series()

        # process sync first => possibly will add new series to track
        new_synced = None
        if is_sync:
            console.status("Fetch followed series from J-Novel Club...")
            follows = await core.fetch_follows(session)
            # new series will also be added to tracked_series
            new_synced, _ = await track.sync_series_forward(
                session, follows, tracked_series, False
            )

            if len(new_synced) == 0:
                console.warning(
                    "There are no new series to sync. Use the [highlight]Follow[/] "
                    "button on a series page on the J-Novel Club website."
                )
                return

        if len(tracked_series) == 0:
            console.warning(
                "There are no tracked series! Use the 'jncep track add' command "
                "first."
            )
            return

        if jnc_url:
            console.status(f"Update '{jnc_url}'...")

            await update.update_url_series(
                session,
                jnc_url,
                epub_generation_options,
                tracked_series,
                new_synced,
                update_options,
            )

        else:
            console.status("Update all series...")

            await update.update_all_series(
                session,
                epub_generation_options,
                tracked_series,
                new_synced,
                update_options,
            )

        # always update and do not notifiy user
        track_manager.write_tracked_series(tracked_series)



from collections import OrderedDict
from functools import partial
import json
import logging
from pathlib import Path

from addict import Dict as Addict
from atomicwrites import atomic_write
import dateutil.parser

from . import config, core, jncweb, spec, utils
from .trio_utils import bag

logger = logging.getLogger(__package__)
console = utils.getConsole()

TRACK_FILE_NAME = "tracked.json"

DEFAULT_CONFIG_FILEPATH = config.config_dir() / TRACK_FILE_NAME

FROM_BEGINNING_CHECK_DATE = "1000-10-10T10:10:10Z"


class TrackConfigManager:
    def __init__(self, config_file_path=None):
        if not config_file_path:
            self.config_file_path = DEFAULT_CONFIG_FILEPATH
        else:
            if config_file_path is Path:
                self.config_file_path = config_file_path
            else:
                self.config_file_path = Path(config_file_path)
            # TODO check read write permission / is file etc...

    # TODO async
    def read_tracked_series(self):
        try:
            with self.config_file_path.open() as json_file:
                # Explicit ordereddict (although should be fine without
                # since Python >= 3.6 dicts are ordered ; spec since 3.7)
                data = json.load(json_file, object_pairs_hook=OrderedDict)
                data = Addict(data)
                return self._convert_to_latest_format(data)
        except FileNotFoundError:
            # first run ?
            return Addict({})

    def write_tracked_series(self, tracked):
        utils.ensure_directory_exists(self.config_file_path.parent)
        with atomic_write(str(self.config_file_path.resolve()), overwrite=True) as f:
            f.write(json.dumps(tracked, sort_keys=True, indent=2))

    def _convert_to_latest_format(self, data):
        converted = {}
        # while at it convert from old format
        # legacy format for tracked parts : just the part instead of object
        # with keys part, name
        # key is slug
        # TODO rename "name" field into "title"
        for series_url_or_slug, value in data.items():
            if not isinstance(value, dict):
                series_slug = series_url_or_slug
                series_url = jncweb.url_from_series_slug(series_slug)
                # low effort way to get some title
                name = series_slug.replace("-", " ").title()
                value = Addict({"name": name, "part": value})
                converted[series_url] = value
            else:
                converted[series_url_or_slug] = value

        converted_b = {}
        for legacy_series_url, value in converted.items():
            new_series_url = jncweb.to_new_website_series_url(legacy_series_url)
            converted_b[new_series_url] = value

        return converted_b


async def track_series(session, tracked_series, series, is_beginning=False):
    parts = core.all_parts_meta(series)

    # record current last part + name
    if not parts or is_beginning:
        # no parts yet, special value 0: processed in update with special case
        pn = 0
        # 0000-... not a valid date so 1111-...
        pdate = "1111-11-11T11:11:11.111Z"

        console.info(
            f"The series '[highlight]{series.raw_data.title}[/]' is now tracked, "
            "starting [highlight]from the beginning[/]",
            style="success",
        )
    else:
        # for the tracking, need the last date: almost always the date of the last part
        # but for some series with volumes released in parallel, can be different
        # see GH #28, also update._update_tracking_data
        # so do additional work for this

        parts = core.all_parts_meta(series)
        last_part_date = max(parts, key=lambda x: x.raw_data.launch)

        last_part_number = parts[-1]
        pn = spec.to_relative_spec_from_part(last_part_number)
        pdate = last_part_date.raw_data.launch

        part_date = dateutil.parser.parse(pdate)
        part_date_formatted = part_date.strftime("%b %d, %Y")
        # TODO display something in case last_part_number and last_part_date_raw do not
        # correspond to the same part?
        console.info(
            f"The series '[highlight]{series.raw_data.title}[/]' is now tracked, "
            f"starting after part [highlight]{pn} [{part_date_formatted}]"
            f"[/]",
            style="success",
        )

    if is_beginning:
        # TODO add a flag to indicate --beginning ; keep track of the last part
        # anyway instead of the special part=0 and last_check_date in that case
        # see also cli.track list_track_series
        # date far in the past: used in case of --use-events => will always end up
        # checking the full metadata of the series (as if --use-events was not used)
        last_check_date = FROM_BEGINNING_CHECK_DATE
    else:
        last_check_date = utils.isoformat_with_z(session.now)

    series_url = jncweb.url_from_series_slug(series.raw_data.slug)
    # TODO class for trackData
    tracked_series[series_url] = Addict(
        {
            "part_date": pdate,
            "part": pn,  # now just for showing to the user in track list
            "name": series.raw_data.title,
            "series_id": series.series_id,
            "last_check_date": last_check_date,
        }
    )


async def sync_series_forward(
    session, follows, tracked_series, is_delete, is_beginning=False
):
    # sync local tracked series based on remote follows
    new_synced = []
    del_synced = []

    async def do_track(jnc_resource):
        series_id = await core.resolve_series(session, jnc_resource)
        series = await core.fetch_meta(session, series_id)
        await track_series(session, tracked_series, series, is_beginning)
        # manga already excluded from follows so no need to check

        series_url = jncweb.url_from_series_slug(series.raw_data.slug)
        new_synced.append(series_url)

    tasks = []
    for jnc_resource in follows:
        if jnc_resource.url in tracked_series:
            continue
        tasks.append(partial(do_track, jnc_resource))

    # result doesn't matter ; just for the exceptions
    await bag(tasks)

    if is_delete:
        followed_index = {f.url: f for f in follows}
        # list() to avoid dictionary changed size during iteration
        for series_url, series_data in list(tracked_series.items()):
            if series_url not in followed_index:
                del tracked_series[series_url]

                console.warning(
                    f"The series '[highlight]{series_data.name}[/]' is no longer "
                    "tracked"
                )

                del_synced.append(series_url)

    return new_synced, del_synced


async def sync_series_backward(session, follows, tracked_series, is_delete):
    # sync remote follows based on locally tracked series
    new_synced = []
    del_synced = []

    async def do_follow(jnc_resource):
        console.info(f"Fetch metadata for '{jnc_resource}'...")
        series_id = await core.resolve_series(session, jnc_resource)
        series = await core.fetch_meta(session, series_id)
        title = series.raw_data.title

        console.info(f"Follow '{title}'...")
        await session.api.follow_series(series.series_id)

        new_synced.append(series_url)

    followed_index = {f.url: f for f in follows}
    tasks = []
    for series_url in tracked_series:
        # series_url is the latest URL format (same as the follows)
        if series_url in followed_index:
            continue
        jnc_resource = jncweb.resource_from_url(series_url)
        tasks.append(partial(do_follow, jnc_resource))

    if is_delete:

        async def do_undollow(jnc_resource):
            # use the follow_raw_data: to avoid another call to the API
            series_id = jnc_resource.follow_raw_data.id
            title = jnc_resource.follow_raw_data.title
            console.warning(f"Unfollow '{title}'...")
            await session.api.unfollow_series(series_id)

            del_synced.append(jnc_resource.url)

        for jnc_resource in follows:
            if jnc_resource.url not in tracked_series:
                tasks.append(partial(do_undollow, jnc_resource))

    # no result needed ; just for the exceptions
    await bag(tasks)

    return new_synced, del_synced



import logging

import click

from . import __version__ as JNCEP_VERSION
from .cli.config import config_manage
from .cli.epub import generate_epub
from .cli.track import track_series
from .cli.update import update_tracked
from .config import apply_options_from_config, DEFAULT_CONFIG_FILEPATH
from .utils import getConsole, module_info, setup_logging

logger = logging.getLogger(module_info())

console = getConsole()


@click.group(
    help="Command-line tool to generate EPUB files for J-Novel Club pre-pub novels"
)
@click.version_option(JNCEP_VERSION, message="v%(version)s")
@click.option(
    "-d",
    "--debug",
    "is_debug",
    is_flag=True,
    help="Flag to activate debug mode",
    required=False,
)
def main(is_debug):
    setup_logging(is_debug)
    try:
        apply_options_from_config()
    except Exception:
        console.warning(
            "There was an error reading the configuration at: "
            f"{DEFAULT_CONFIG_FILEPATH}. Continuing..."
        )


main.add_command(generate_epub)
main.add_command(track_series)
main.add_command(update_tracked)
main.add_command(config_manage)

if __name__ == "__main__":
    main()



import logging
import re

import attr

logger = logging.getLogger(__name__)

RANGE_SEP = ":"

SERIES = "SERIES_ALL"
VOLUME = "VOLUME_ALL"
PART = "SINGLE_PART"
START_OF_VOLUME = "START_OF_VOLUME"
END_OF_VOLUME = "END_OF_VOLUME"
START_OF_SERIES = "START_OF_SERIES"
END_OF_SERIES = "END_OF_SERIES"


class SpecError(Exception):
    pass


@attr.s
class Single:
    type_ = attr.ib()
    spec = attr.ib()

    def has_volume(self, volume) -> bool:
        if self.type_ == SERIES:
            return True
        elif self.type_ == VOLUME:
            return self.spec == volume.num
        # part
        vn, _ = self.spec
        return volume.num == vn

    def has_part(self, part) -> bool:
        if self.type_ == SERIES:
            return True

        if not self.has_volume(part.volume):
            return False
        if self.type_ == VOLUME:
            return True

        _, pn = self.spec
        return part.num_in_volume == pn

    def normalize_and_verify(self, series):
        if self.type_ == SERIES:
            return
        elif self.type_ == VOLUME:
            original_spec = self.spec
            self.spec = _normalize_volume(self.spec, series)
            if 1 <= self.spec <= len(series.volumes):
                return
            raise SpecError(f"Bad spec for series: Volume '{original_spec}' not found")
        # part
        vn, pn = self.spec
        original_spec = self.spec

        vn = _normalize_volume(vn, series)
        self.spec = (vn, pn)
        if vn < 1 or vn > len(series.volumes):
            raise SpecError(
                f"Bad spec for series: Volume '{original_spec[0]}' not found"
            )

        volume = series.volumes[vn - 1]
        if 1 <= pn <= len(volume.parts):
            return
        raise SpecError(
            f"Bad spec for series: Part '{original_spec[0]}.{original_spec[1]}' "
            "not found"
        )


@attr.s
class Interval:
    start = attr.ib()
    end = attr.ib()

    def has_volume(self, volume) -> bool:
        if self.start == START_OF_SERIES:
            # spec is a tuple (START_OF / END_OF cannot happen together)
            # if analyse_spec is used
            vn2, _ = self.end.spec
            return volume.num <= vn2

        vn, _ = self.start.spec
        if self.end == END_OF_SERIES:
            return volume.num >= vn

        vn2, _ = self.end.spec
        return vn <= volume.num <= vn2

    def has_part(self, part) -> bool:
        if self.start == START_OF_SERIES:
            # spec is a tuple (see comment in has_volume)
            vn2, pn2 = self.end.spec
            if part.volume.num < vn2:
                return True

            if part.volume.num > vn2:
                return False

            # same volume
            if pn2 == END_OF_VOLUME:
                return True

            return part.num_in_volume <= pn2

        # spec is a tuple
        vn, pn = self.start.spec
        if self.end == END_OF_SERIES:
            if part.volume.num > vn:
                return True

            if part.volume.num < vn:
                return False

            # same volume
            if pn == START_OF_VOLUME:
                return True

            return part.num_in_volume >= pn

        vn2, pn2 = self.end.spec
        if vn < part.volume.num < vn2:
            return True

        if part.volume.num < vn or part.volume.num > vn2:
            return False

        if vn == vn2:
            if pn == START_OF_VOLUME:
                if pn2 == END_OF_VOLUME:
                    return True

                return part.num_in_volume <= pn2
            else:
                if pn2 == END_OF_VOLUME:
                    return part.num_in_volume >= pn

                return pn <= part.num_in_volume <= pn2
        else:
            if part.volume.num == vn:
                if pn == START_OF_VOLUME:
                    return True

                return part.num_in_volume >= pn
            else:
                assert part.volume.num == vn2
                if pn2 == END_OF_VOLUME:
                    return True

                return part.num_in_volume <= pn2

    def normalize_and_verify(self, series):
        if self.start != START_OF_SERIES:
            vn, pn = self.start.spec
            original_spec1 = self.start.spec
            vn = _normalize_volume(vn, series)
            self.start.spec = (vn, pn)

            if vn < 1 or vn > len(series.volumes):
                raise SpecError(
                    f"Bad left spec for series: Volume '{original_spec1[0]}' not found"
                )

            if pn != START_OF_VOLUME:
                volume = series.volumes[vn - 1]
                if pn < 1 or pn > len(volume.parts):
                    raise SpecError(
                        "Bad left spec for series: Part "
                        f"'{original_spec1[0]}.{original_spec1[1]}' "
                        "not found"
                    )

        if self.end != END_OF_SERIES:
            vn, pn = self.end.spec
            original_spec2 = self.end.spec
            vn = _normalize_volume(vn, series)
            self.end.spec = (vn, pn)

            if vn < 1 or vn > len(series.volumes):
                raise SpecError(
                    f"Bad right spec for series: Volume '{original_spec2[0]}' "
                    "not found"
                )

            if pn != END_OF_VOLUME:
                volume = series.volumes[vn - 1]
                if pn < 1 or pn > len(volume.parts):
                    raise SpecError(
                        "Bad right spec for series: Part "
                        f"'{original_spec2[0]}.{original_spec2[1]}' "
                        "not found"
                    )

        # already tested that both sides are valid in the series
        # test if left <= right
        if self.start != START_OF_SERIES and self.end != END_OF_SERIES:
            vn1, pn1 = self.start.spec
            vn2, pn2 = self.end.spec

            if vn2 > vn1:
                return

            if vn2 < vn1:
                raise SpecError(
                    f"Bad spec for series: Volume '{original_spec2[0]}' is before "
                    f"Volume '{original_spec1[0]}'"
                )

            # vn1 == vn2:
            if pn1 != START_OF_VOLUME and pn2 != END_OF_VOLUME:
                if pn2 < pn1:
                    raise SpecError(
                        "Bad spec for series: "
                        f"Part '{original_spec2[0]}.{original_spec2[1]}' "
                        "is before "
                        f"Part '{original_spec1[0]}.{original_spec1[1]}'"
                    )


def _normalize_volume(vn, series):
    # handle negative volume numbers
    if vn < 0:
        # numerotation of volumes in spec start at 1
        return len(series.volumes) + vn + 1
    return vn


@attr.s
class IdentifierSpec:
    # not really a spec (part:part) => represents a requests for series, vol or part by
    # id
    type_ = attr.ib()
    volume_id = attr.ib(None)
    part_id = attr.ib(None)

    def has_volume(self, volume) -> bool:
        # assumes : only check a single series with the spec
        if self.type_ == SERIES:
            return True

        return self.volume_id == volume.volume_id

    def has_part(self, part) -> bool:
        if self.type_ == SERIES:
            return True

        if self.type_ == VOLUME:
            return self.volume_id == part.volume.volume_id

        return self.part_id == part.part_id


def to_relative_spec_from_part(part):
    volume_number = part.volume.num
    part_number = part.num_in_volume
    return f"{volume_number}.{part_number}"


def analyze_part_specs(part_specs):
    """v(.p):v2(.p) or v(.p): or :v(.p) or v(.p) or :"""

    part_specs = part_specs.strip()

    if part_specs == RANGE_SEP:
        return Single(SERIES, None)

    return _analyze_volume_part_specs(part_specs)


def _analyze_volume_part_specs(part_specs):
    sides = part_specs.split(RANGE_SEP)
    if len(sides) > 2:
        raise ValueError("Multiple ':' in part specs")

    reg = r"^\s*(-?\d+)(?:\.(\d+))?\s*$"
    if len(sides) == 1:
        # not a range: single part
        m = re.match(reg, sides[0])
        if not m:
            raise ValueError(
                "Specification must be a of the form 'vol[.part]' (part is optional)"
            )
        fv = int(m.group(1))
        if m.group(2):
            # only the part specified
            fp = int(m.group(2))
            return Single(PART, (fv, fp))
        else:
            # full volume
            return Single(VOLUME, fv)

    # range
    m1 = re.match(reg, sides[0])
    m2 = re.match(reg, sides[1])
    if (
        (not m1 and not m2)
        # left side not valid
        or (not m1 and len(sides[0]) > 0)
        # right side not valid
        or (not m2 and len(sides[1]) > 0)
    ):
        msg = (
            "Part specification must be vol[.part]:vol[.part] or vol[.part]: or "
            ":vol[.part]"
        )
        raise ValueError(msg)

    if m1:
        fv = int(m1.group(1))
        if m1.group(2):
            fp = int(m1.group(2))
        else:
            fp = START_OF_VOLUME
        start = Single(PART, (fv, fp))
    else:
        start = START_OF_SERIES

    if m2:
        lv = int(m2.group(1))
        if m2.group(2):
            lp = int(m2.group(2))
        else:
            lp = END_OF_VOLUME
        end = Single(PART, (lv, lp))
    else:
        end = END_OF_SERIES

    # both sides
    return Interval(start, end)



from collections import namedtuple
from functools import partial
import logging
import sys

import attr
import dateutil
import trio

from . import core, jncweb, spec, utils
from .trio_utils import bag

logger = logging.getLogger(__package__)
console = utils.getConsole()


@attr.s
class UpdateResult:
    series = attr.ib(None)
    is_error = attr.ib(False)
    is_updated = attr.ib(None)
    is_considered = attr.ib(True)
    # to indicate a series with expired parts only
    # will set to latest part or will always have error
    # if stalled
    is_force_set_updated = attr.ib(False)
    is_update_last_checked = attr.ib(True)
    parts_downloaded = attr.ib(None)


UpdateOptions = namedtuple(
    "UpdateOptions",
    [
        "is_sync",
        "is_whole_volume",
        "is_whole_volume_on_final_part",
        "is_use_events",
    ],
)


async def update_url_series(
    session,
    jnc_url,
    epub_generation_options,
    tracked_series,
    new_synced,
    update_options,
):
    # for single url => if error no catch : let it crash and report to the user
    jnc_resource = jncweb.resource_from_url(jnc_url)
    series_id = await core.resolve_series(session, jnc_resource)
    series = await core.fetch_meta(session, series_id)

    series_url = jncweb.url_from_series_slug(series.raw_data.slug)

    if series_url not in tracked_series:
        console.warning(
            f"The series '[highlight]{series.raw_data.title}[/]' is not tracked! "
            f"Use the 'jncep track add' command first."
        )
        return

    if update_options.is_sync:
        # not very useful but make it possible
        # only consider newly synced series if --sync used
        # to mirror case with no URL argument
        if series_url not in new_synced:
            console.warning(
                f"The series '[highlight]{series.raw_data.title}[/]' is not "
                "among the tracked series added from syncing. Use 'jncep update' "
                "without --sync."
            )
            return

    series_details = tracked_series[series_url]

    is_need_check = True
    is_check_events = update_options.is_use_events and _can_use_events_feed(
        series_details
    )
    if is_check_events:
        console.status("Checking J-Novel Club events feed...", clear=False)
        start_date = series_details.last_check_date
        events = await core.fetch_events(session, start_date)
        is_need_check = _verify_series_needs_update_check(events, series_details)
        if not is_need_check:
            update_result = UpdateResult(is_updated=False)
        console.pop_status()

    if is_need_check:
        # the series has just been synced so force EPUB gen from start in this case
        is_force_from_beginning = update_options.is_sync
        update_result = await _create_epub_for_new_parts(
            session,
            series_details,
            series,
            epub_generation_options,
            update_options,
            is_force_from_beginning,
        )

    if update_result.is_updated:
        emoji = ""
        if console.is_advanced():
            emoji = "\u2714 "
        console.info(
            f"{emoji}The series '[highlight]{series.raw_data.title}[/]' has "
            "been updated!",
            style="success",
        )
    else:
        console.info(
            f"The series '[highlight]{series.raw_data.title}[/]' is already up "
            "to date!",
            style="success",
        )
        if is_check_events and is_need_check:
            # events feed said the series was updated but checking the series says
            # there was no update => incoherent : for now, do not update the checl
            # date
            # possible also if tracked.json updated manually : last_checked_date is
            # before the date of the last downloaded part in the file
            update_result.is_update_last_checked = False

    _update_tracking_data(series_details, series, update_result, session.now)


async def update_all_series(
    session,
    epub_generation_options,
    tracked_series,
    new_synced,
    update_options,
):
    # is_sync: all parts from beginning so no need for the events
    if (
        not update_options.is_sync
        and update_options.is_use_events
        and _can_any_use_events_feed(tracked_series)
    ):
        console.status("Checking J-Novel Club events feed...", clear=False)
        start_date = _min_last_check_date(tracked_series)
        events = await core.fetch_events(session, start_date)
        console.pop_status()
    else:
        events = None

    series_details_a = []
    tasks = []
    for series_url, series_details in tracked_series.items():
        tasks.append(
            partial(
                _handle_series,
                session,
                series_url,
                series_details,
                epub_generation_options,
                update_options.is_sync,
                new_synced,
                update_options,
                events,
            )
        )
        series_details_a.append(series_details)

    results = await bag(tasks)

    num_updated = 0
    num_errors = 0
    update_result: UpdateResult
    for i, update_result in enumerate(results):
        series_details = series_details_a[i]

        # --sync has bee used and series is not part of the synced series so
        # has not been checked
        if not update_result.is_considered:
            continue

        if update_result.is_updated:
            num_updated += 1

        if update_result.is_error:
            num_errors += 1
        else:
            # the update of tracking has some conditions besides
            # just the series updated
            _update_tracking_data(
                series_details, update_result.series, update_result, session.now
            )

    if num_errors > 0:
        console.error("Some series could not be updated!")

    emoji = ""
    if console.is_advanced():
        emoji = "\u2728 "

    if num_updated == 0 and num_errors == 0:
        # second clause => all in error
        console.info(
            f"{emoji}All series are already up to date!",
            style="success",
        )

    if num_updated > 0:
        console.info(
            f"{emoji}{num_updated} series sucessfully updated!",
            style="success",
        )


def _update_tracking_data(series_details, series_meta, update_result, check_date):
    # alway update this : in case --use-events is used
    if update_result.is_update_last_checked:
        series_details.last_check_date = utils.isoformat_with_z(check_date)

    # not always available (if series not checked for example)
    if series_meta:
        # should stay always the same
        series_details.series_id = series_meta.series_id

    if not (update_result.is_updated or update_result.is_force_set_updated):
        return

    parts = core.all_parts_meta(series_meta)
    assert bool(parts)

    # this is only used for display to the user in track list
    last_part_number = parts[-1]
    pn = spec.to_relative_spec_from_part(last_part_number)
    series_details.part = pn

    # if series has more than one volume in parallel, it can happen that the last
    # part (in number) is released before (in date) a part that comes before (in number)
    # we use the part_date for the update process => last part (in date) for this
    # instead of using the date of the last part (in number)
    last_part_date = max(parts, key=lambda x: x.raw_data.launch)
    pdate = last_part_date.raw_data.launch

    series_details.part_date = pdate


async def _handle_series(
    session,
    series_url,
    series_details,
    epub_generation_options,
    is_sync,
    new_synced,
    update_options,
    events,
):
    series = None
    try:
        if is_sync and series_url not in new_synced:
            return UpdateResult(is_considered=False)

        is_need_check = False
        is_check_events = events and _can_use_events_feed(series_details)
        if is_check_events:
            is_need_check = _verify_series_needs_update_check(events, series_details)
            if not is_need_check:
                return UpdateResult(is_updated=False)
            # else the standard check continues

        jnc_resource = jncweb.resource_from_url(series_url)
        series_id = await core.resolve_series(session, jnc_resource)
        series = await core.fetch_meta(session, series_id)

        # generate from the start if the series is newly sync
        is_force_from_beginning = is_sync
        update_result = await _create_epub_for_new_parts(
            session,
            series_details,
            series,
            epub_generation_options,
            update_options,
            is_force_from_beginning,
        )

        if update_result.is_updated:
            emoji = ""
            if console.is_advanced():
                emoji = "\u2714 "
            console.info(
                f"{emoji}The series '[highlight]{series.raw_data.title}[/]' has "
                "been updated!",
                style="success",
            )
        else:
            if is_check_events and is_need_check:
                # incoherence between feed and series data
                # assumes maybe advertised part has not been released yet (but will be)
                # so do not advance check_date (so next check, if the part is released
                # it will be picked up)
                update_result.is_update_last_checked = False

        return update_result

    except (trio.MultiError, Exception) as ex:
        if series and series.raw_data:
            title = series.raw_data.title
        else:
            title = series_url

        emoji = ""
        if console.is_advanced():
            emoji = "\u274C "
        # FIXME show the user some feedback as to the nature of the error
        console.error(
            f"{emoji}Error updating '{title}'! "
            "(run 'jncep -d update' for more details)",
        )
        logger.debug(f"Error _handle_series: {ex}", exc_info=sys.exc_info())
        # series_meta may be None if error during retrieval
        return UpdateResult(is_error=True)


def _verify_series_needs_update_check(event_feed, series_details):
    last_check_date = dateutil.parser.parse(series_details.last_check_date)

    events, has_reached_limit, first_event_date = event_feed

    # shortcuts for some special cases

    if len(events) == 0:
        # events only contain the necessary events for dates between last_check and now:
        # no events => no update
        return False

    # last_check_date is specific to the series; but event feed is checked taking into
    # account all series so event if has_reached_limit is True, series may not need to
    # be checked
    if has_reached_limit and last_check_date <= first_event_date:
        # events doesn't go far enough in the past: Not possible to know for sure
        # if there are no updates
        # assumes need check
        return True

    series_id = series_details.series_id

    for event in events:
        if "details" in event and event.details.startswith("Release of Part"):
            # no s in JNC attr
            series = event.serie
            if series.id != series_id:
                continue

            launch_date = dateutil.parser.parse(event.launch)
            # <= : last_check_date is the session.now of the previous check so if
            # equal to last_check_date, already included in previous check
            # see core.fetch_events request parameters
            if launch_date <= last_check_date:
                # the events are ordered by launch desc so can never be false after
                break

            # return only that there has been updates
            # the standard check for the series will be done after
            # TODO return specific parts ?
            return True

    return False


def _can_any_use_events_feed(tracking_data):
    return any(
        (
            _can_use_events_feed(series_details)
            for series_details in tracking_data.values()
        )
    )


def _min_last_check_date(tracking_data):
    # all dates are encoded in the same ISO format
    check_dates = (
        d.last_check_date for d in tracking_data.values() if _can_use_events_feed(d)
    )
    return min(check_dates)


def _can_use_events_feed(series_details):
    # the 2 attributes have been added later so may not always be there
    return "series_id" in series_details and "last_check_date" in series_details


async def _create_epub_for_new_parts(
    session,
    series_details,
    series,
    epub_generation_options,
    update_options,
    is_force_from_beginning=False,
):
    parts = core.all_parts_meta(series)

    if series_details.part == 0 or is_force_from_beginning:
        # Firt clause: special processing : means there was no part available when the
        # series was started tracking

        update_result = await _update_from_beginning(
            session, series, parts, epub_generation_options
        )
        # no need for _generate_whole_volume_on_final_part: if from the beginning +
        # final included => the whole volume is already generated
        return update_result
    else:
        update_result = await _update_new_parts(
            session,
            series_details,
            series,
            parts,
            epub_generation_options,
            update_options,
        )

        if (
            update_result.is_updated
            and update_options.is_whole_volume_on_final_part
            and not update_options.is_whole_volume
        ):
            # TODO do that as soon as we know which parts to download before
            # fill_covers_and_content
            parts_downloaded = update_result.parts_downloaded
            await _generate_whole_volume_on_final_part(
                session, series, parts_downloaded, epub_generation_options
            )

        return update_result


async def _update_from_beginning(session, series, parts, epub_generation_options):
    # still no part ?
    if not parts:
        return UpdateResult(is_updated=False)
    else:
        console.info(
            f"The series '[highlight]{series.raw_data.title}[/]' " "will be updated..."
        )

        part_filter = partial(core.is_part_available, session.now)

        (
            volumes_to_download,
            parts_to_download,
        ) = core.relevant_volumes_and_parts_for_content(series, part_filter)
        volumes_for_cover = core.relevant_volumes_for_cover(
            volumes_to_download, epub_generation_options.is_by_volume
        )

        await core.fill_covers_and_content(
            session, volumes_for_cover, parts_to_download
        )
        await core.create_epub(
            series,
            volumes_to_download,
            parts_to_download,
            epub_generation_options,
        )

        return UpdateResult(series, is_updated=True, parts_downloaded=parts_to_download)


async def _update_new_parts(
    session,
    series_details,
    series,
    parts,
    epub_generation_options,
    update_options,
):
    if not series_details.part_date:
        # if here => old format, first lookup date of last part and use that
        # still useful for stalled series so keep it
        part_spec = spec.analyze_part_specs(series_details.part)
        for part in parts:
            if part_spec.has_part(part):
                # will be filled if the part still exists (it should)
                # TODO case it doesn't ? eg tracked.json filled by hand
                last_update_part = part
                break
        # in UTC
        last_update_date = last_update_part.raw_data.launch
    else:
        # new format : date is recorded
        last_update_date = series_details.part_date

    last_update_date = dateutil.parser.parse(last_update_date)

    parts_release_after_date = _filter_parts_released_after_date(
        last_update_date, parts
    )

    if not parts_release_after_date:
        # not updated
        return UpdateResult(series, is_updated=False)

    available_parts_to_download = [
        part
        for part in parts_release_after_date
        if core.is_part_available(session.now, part)
    ]

    if not available_parts_to_download:
        console.warning(
            f"All updated parts for '[highlight]{series.raw_data.title}[/]' "
            "have expired!"
        )
        # not updated but the series will still have its tracking data changed
        # in tracking config ; if not, the message above will always be displayed
        # should be rare (if updating often) ; also first part is preview
        # so even rarer
        return UpdateResult(series=series, is_updated=False, is_force_set_updated=True)

    console.info(
        f"The series '[highlight]{series.raw_data.title}[/]' will be updated..."
    )

    # after the to update notification
    if len(available_parts_to_download) != len(parts_release_after_date):
        console.warning(
            f"Some parts for '[highlight]{series.raw_data.title}[/]' have " "expired!"
        )

    parts_id_to_download = set((part.part_id for part in available_parts_to_download))

    # availability alread tested
    def simple_part_filter(part):
        return part.part_id in parts_id_to_download

    (
        volumes_to_download,
        parts_to_download,
    ) = core.relevant_volumes_and_parts_for_content(series, simple_part_filter)

    if update_options.is_whole_volume:
        # second pass : filter on the volumes_to_download
        # all the parts of those volumes must be downloaded
        volumes_id_to_download = set((v.volume_id for v in volumes_to_download))

        def whole_volume_part_filter(part):
            return (
                part.volume.volume_id in volumes_id_to_download
                and core.is_part_available(session.now, part)
            )

        (
            volumes_to_download,
            parts_to_download,
        ) = core.relevant_volumes_and_parts_for_content(
            series, whole_volume_part_filter
        )

    volumes_for_cover = core.relevant_volumes_for_cover(
        volumes_to_download, epub_generation_options.is_by_volume
    )

    await core.fill_covers_and_content(session, volumes_for_cover, parts_to_download)

    await core.create_epub(
        series,
        volumes_to_download,
        parts_to_download,
        epub_generation_options,
    )

    return UpdateResult(series, is_updated=True, parts_downloaded=parts_to_download)


def _is_released_after_date(date, part_date_s):
    # all date strings are in ISO format
    # so no need to parse really
    # parsing just to be safe
    # in case different shape like ms part or not (which throws str comp off)
    launch_date = dateutil.parser.parse(part_date_s)
    return launch_date > date


def _filter_parts_released_after_date(date, parts):
    parts_to_download = []
    for part in parts:
        if _is_released_after_date(date, part.raw_data.launch):
            parts_to_download.append(part)

    return parts_to_download


async def _generate_whole_volume_on_final_part(
    session, series, parts_downloaded, epub_generation_options
):
    # check if any part included in the update is the final part of its volume
    for part in parts_downloaded:
        # only max one part can be final in a volume
        if not core._is_part_final(part):
            continue

        # check if possibly all parts have already been downloaded as part of the
        # update
        for volpart in part.volume.parts:
            if volpart not in parts_downloaded:
                break
        else:
            # all the parts have been downloaded in the normal course
            # of things, so we skip regenerating the whole volume
            continue

        console.info(
            "The complete volume "
            f"'[highlight]{part.volume.raw_data.title}[/]' will be "
            "downloaded..."
        )

        # With JNC if the final part can be downloaded, the rest of the
        # volume is also available for download so no need to check

        await core.fill_covers_and_content(session, [part.volume], part.volume.parts)
        await core.create_epub(
            series,
            [part.volume],
            part.volume.parts,
            epub_generation_options,
        )



from addict import Dict as Addict
import pytest

from jncep.model import Part, Volume
from jncep.spec import analyze_part_specs, SpecError


def _to_vp(vn, pn):
    volume = Volume(None, None, vn)
    part = Part(None, None, pn, volume)

    return volume, part


def _to_series(num_parts):
    series = Addict(
        {
            "volumes": [
                {"num": i, "parts": [None] * pn} for i, pn in enumerate(num_parts)
            ]
        }
    )
    return series


def test_open_end_no_part():
    spec = analyze_part_specs("2:")

    volume, part = _to_vp(1, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(2, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_open_end_with_part():
    spec = analyze_part_specs("2.4:")

    volume, part = _to_vp(1, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(2, 1)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(2, 4)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(3, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_open_start_no_part():
    spec = analyze_part_specs(":6")

    volume, part = _to_vp(2, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(6, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(7, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)


def test_open_start_with_part():
    spec = analyze_part_specs(":6.4")

    volume, part = _to_vp(2, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(6, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(6, 7)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(7, 3)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)


def test_whole_series():
    spec = analyze_part_specs(":")

    volume, part = _to_vp(6, 3)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(1, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_same_vol_diff_parts():
    spec = analyze_part_specs("9.6:9.8")

    volume, part = _to_vp(6, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(9, 9)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(9, 4)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(9, 6)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_same_parts():
    spec = analyze_part_specs("9.6:9.6")

    volume, part = _to_vp(9, 4)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(9, 6)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_diff_vol():
    spec = analyze_part_specs("5.6:7.8")

    volume, part = _to_vp(4, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(5, 4)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(5, 6)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(5, 9)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(6, 4)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(7, 9)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(7, 8)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(7, 4)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_diff_parts_negative_volumes():
    series = _to_series([8, 11, 10, 7, 12])
    spec = analyze_part_specs("-1.3:-1.6")
    spec.normalize_and_verify(series)

    volume, part = _to_vp(4, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(5, 4)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    spec = analyze_part_specs("-5.3:-2.6")
    spec.normalize_and_verify(series)

    volume, part = _to_vp(5, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(2, 4)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    spec = analyze_part_specs("-4.3:3.6")
    spec.normalize_and_verify(series)

    volume, part = _to_vp(2, 4)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(3, 4)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(2, 1)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(4, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)


def test_single_volume():
    spec = analyze_part_specs("9")

    volume, part = _to_vp(6, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(9, 9)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_single_negative_volume():
    series = _to_series([8, 11, 10])

    spec = analyze_part_specs("-2")
    # necessary for negative volumes
    spec.normalize_and_verify(series)

    volume, part = _to_vp(1, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(2, 9)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(3, 2)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)


def test_single_part():
    spec = analyze_part_specs("6.7")

    volume, part = _to_vp(6, 1)
    assert spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(6, 7)
    assert spec.has_volume(volume)
    assert spec.has_part(part)


def test_single_part_negative_volume():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs("-2.1")
    spec.normalize_and_verify(series)

    volume, part = _to_vp(2, 1)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)

    volume, part = _to_vp(3, 1)
    assert spec.has_volume(volume)
    assert spec.has_part(part)

    volume, part = _to_vp(4, 2)
    assert not spec.has_volume(volume)
    assert not spec.has_part(part)


def test_verifiy_whole_series():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs(":")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")


def test_verify_open_end_no_part():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs("2:")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("5:")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-5:")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-4:")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")


def test_verify_open_end_with_part():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs("2.4:")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("5.3:")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("2.12:")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-5.3:")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-3.12:")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-4.8:")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")


def test_verify_open_start_no_part():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs(":1")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs(":6")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs(":-5")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs(":-4")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")


def test_verify_open_start_with_part():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs(":4.4")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs(":4.10")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs(":5.2")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs(":-5.2")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs(":-1.4")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs(":-1.10")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)


def test_verify_same_vol_diff_parts():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs("-5.8:2.6")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("4.6:4.8")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("4.8:4.6")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-1.6:4.8")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("-1.8:4.6")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)


def test_verify_same_parts():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs("3.6:3.6")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("-1.8:4.8")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("-5.6:-5.6")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)


def test_verify_diff_vol():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs("5.6:7.8")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("3.2:4.9")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("2.2:4")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("2:4.5")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("4.5:2")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("2.12:4.7")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("2.2:1.7")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("2.1:4.10")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-5.1:4.10")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("1.3:-1.10")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("2.3:-1")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")


def test_verify_single_volume():
    series = _to_series([8, 11, 10])

    spec = analyze_part_specs("3")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("4")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-2")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("-5")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)


def test_verify_single_part():
    series = _to_series([8, 11, 10, 9])

    spec = analyze_part_specs("3.1")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("3.11")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("5.6")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-2.1")
    try:
        spec.normalize_and_verify(series)
    except SpecError:
        pytest.fail("Unexpected")

    spec = analyze_part_specs("-2.12")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)

    spec = analyze_part_specs("-5")
    with pytest.raises(SpecError):
        spec.normalize_and_verify(series)



import shutil

import click

from .. import config, track, utils
from .base import CatchAllExceptionsCommand

console = utils.getConsole()


@click.group(name="config", help="Manage configuration")
def config_manage():
    pass


@config_manage.command(
    name="show", help="List configuration details", cls=CatchAllExceptionsCommand
)
def config_list():
    config_dir = config.config_dir()
    if not config_dir.exists():
        console.warning("No configuration folder found!")
        console.info(f"The recommended location is: [highlight]{config_dir}[/]")
        return

    if not config_dir.is_dir:
        console.warning(f"Not a folder: [highlight]{config_dir}[/]")

    console.info(f"Configuration folder: [highlight]{config_dir}[/]")
    files = list(config_dir.iterdir())
    for f_ in files:
        if f_.is_file():
            if f_.name == track.TRACK_FILE_NAME:
                console.info(f"Found tracking file: [highlight]{f_.name}[/]")
                _track_file_summary(f_)
                continue
            if f_.name == config.CONFIG_FILE_NAME:
                console.info(f"Found config file: [highlight]{f_.name}[/]")
                _config_file_summary(f_)
                continue
        # ignore everything else


def _track_file_summary(file_path):
    track_config_manager = track.TrackConfigManager(file_path)
    tracked_series = track_config_manager.read_tracked_series()
    len_ts = len(tracked_series)
    console.info(f"{len_ts} series tracked")


def _config_file_summary(file_path):
    config_manager = config.ConfigManager(file_path)
    config_options = config_manager.read_config_options()
    # TOP_SECTION is always there (default section)
    jncep_s = config_options[config.TOP_SECTION]
    if len(jncep_s) == 0:
        console.info("No option set")

    allowed_options = config.list_available_config_options()
    for option in jncep_s:
        # ignore others that are unknown to JNCEP
        if option not in allowed_options:
            continue
        console.info(f"Option: [highlight]{option}[/] => [green]{jncep_s[option]}[/]")


@config_manage.command(
    name="list", help="List configuration options", cls=CatchAllExceptionsCommand
)
def list_options():
    options = config.list_available_config_options()

    rows = [(f"[highlight]{o}[/]", h) for o, h in options.items()]

    # option column: adjust with max margin of 5
    max_len = 0
    for option in options:
        max_len = max(max_len, len(option))
    column_width = max_len + 5

    console.info_table(rows, maxcolwidths=(column_width, 40))


def _wrap(wrapper, text):
    return "\n".join(wrapper.wrap(text))


@config_manage.command(
    name="set", help="Set configuration option", cls=CatchAllExceptionsCommand
)
@click.argument("option", metavar="OPTION", required=True)
@click.argument("value", metavar="VALUE", required=True)
def set_option(option, value):
    config_manager = config.ConfigManager()
    config_options = config_manager.read_config_options()

    option = config.set_config_option(config_options, option, value)
    console.info(
        f"Option '[highlight]{option}[/]' set to '[highlight]{value}[/]'",
        style="success",
    )

    is_needs_created = not config_manager.config_file_path.exists()
    # if config file didn't exist, this will create it
    config_manager.write_config_options(config_options)
    if is_needs_created:
        # mention the creation to user
        console.info(
            f"Config file created at: [highlight]{config_manager.config_file_path}[/]"
        )


@config_manage.command(
    name="unset", help="Delete configuration option", cls=CatchAllExceptionsCommand
)
@click.argument("option", metavar="OPTION", required=True)
def unset_option(option):
    config_manager = config.ConfigManager()

    if not config_manager.config_file_path.exists():
        console.warning(
            f"No config file found at: [highlight]{config_manager.config_file_path}[/]"
        )
        return

    config_options = config_manager.read_config_options()

    option, is_deleted = config.unset_config_option(config_options, option)
    if not is_deleted:
        console.warning(f"Option '[highlight]{option}[/]' not set in config")
        return
    else:
        console.info(f"Option '[highlight]{option}[/]' unset", style="success")

    config_manager.write_config_options(config_options)


@config_manage.command(
    name="init", help="Create configuration file", cls=CatchAllExceptionsCommand
)
def init_config():
    config_filepath = config.DEFAULT_CONFIG_FILEPATH
    if config_filepath.exists():
        console.warning(f"Config file already exists: [highlight]{config_filepath}[/]")
        return

    config_manager = config.ConfigManager(config_filepath)
    # will create empty config file
    config_options = config_manager.read_config_options()
    config_manager.write_config_options(config_options)

    console.info(
        f"New empty config file created: [highlight]{config_filepath}[/]",
        style="success",
    )


@config_manage.command(
    name="migrate",
    help=f"Migrate to standard configuration folder [{config.APPDATA_CONFIG_DIR}]",
    cls=CatchAllExceptionsCommand,
)
def config_migrate():
    if not config.has_config_dir():
        console.warning("No configuration folder found!")
        console.info(
            f"The recommended location is: [highlight]{config.APPDATA_CONFIG_DIR}[/]"
        )
        return

    current_config_dir = config.config_dir()
    if current_config_dir == config.APPDATA_CONFIG_DIR:
        console.warning(
            f"Configuration is already in: [highlight]{config.APPDATA_CONFIG_DIR}[/]"
        )
        return

    migrate_config_dir = config.APPDATA_CONFIG_DIR

    # configuration is in legacy_config_dir => migrate to appdir

    migrate_config_dir.mkdir(parents=True)

    from_track_filepath = current_config_dir / track.TRACK_FILE_NAME
    if from_track_filepath.exists():
        to_track_filepath = migrate_config_dir / track.TRACK_FILE_NAME
        shutil.copy2(from_track_filepath, to_track_filepath)

    from_config_filepath = current_config_dir / config.CONFIG_FILE_NAME
    if from_config_filepath.exists():
        to_config_filepath = migrate_config_dir / config.CONFIG_FILE_NAME
        shutil.copy2(from_config_filepath, to_config_filepath)

    console.info(
        "[success]"
        f"The configuration is now in: [highlight]{migrate_config_dir}[/]"
        "[/]\n"
        f"You may delete: [highlight]{current_config_dir}[/]"
    )



from collections import deque
from datetime import timezone
import inspect
import logging
from pathlib import Path
import re
import sys
import unicodedata

from addict import Dict as Addict
import dateutil.parser
import rich.console
from rich.table import Column, Table
import rich.theme

logger = logging.getLogger(__name__)


def setup_logging(is_debug):
    format = "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"
    if not logging.getLogger().handlers:
        # coloredlogs is not used
        logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=format)
    else:
        # Colored_logs has installed its own handler but not configured the way
        # I want
        logging.getLogger().handlers[0].formatter.datefmt = "%Y-%m-%d,%H:%M:%S,%f"
        # coloredlogs changes the level of the handler
        logging.getLogger().handlers[0].setLevel(logging.NOTSET)

    logger_pkg = logging.getLogger(__package__)
    if is_debug:
        logger_pkg.setLevel(logging.DEBUG)
        # keep debug console (ie logging)
        # or issues mixing the print and the logs with Rich
    else:
        logger_pkg.setLevel(logging.INFO)
        getConsole().console = RichConsole()


def is_debug():
    logger_pkg = logging.getLogger(__package__)
    return logger_pkg.level <= logging.DEBUG


def tryint(val):
    try:
        return int(val)
    except Exception:
        return None


def to_yn(b):
    return "yes" if b else "no"


def to_safe_filename(name):
    name = "".join(
        c for c in unicodedata.normalize("NFD", name) if unicodedata.category(c) != "Mn"
    )
    safe = re.sub(r"[^0-9a-zA-Z_]+", "_", name)
    safe = safe.strip("_")
    return safe


def module_info():
    # for main module : its __name__ is __main__
    # so find out its real name
    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    return mod.__spec__.name


def isoformat_with_z(d):
    # like the date format used by the JNC API
    if d.tzinfo != timezone.utc:
        # just in case but shouldn't happen : only date in UTC is used in the code
        raise ValueError("Date not in UTC!")
    return d.replace(microsecond=0).isoformat().replace("+00:00", "Z")


def compare_date_isoformat(d1, d2):
    # convert in case ms are used
    date1 = dateutil.parser.parse(d1)
    date2 = dateutil.parser.parse(d2)

    if date1 == date2:
        return 0
    elif date1 < date2:
        return -1
    return 1


def ensure_directory_exists(path):
    Path(path).mkdir(parents=True, exist_ok=True)


def deep_freeze(data):
    if type(data) is Addict:
        data.freeze()
        for value in data.values():
            if type(value) is list:
                for v in value:
                    deep_freeze(v)


rich_theme = rich.theme.Theme(
    {
        "success": "green",
        "warning": "yellow",
        "error": "red",
        "highlight": "magenta",
        "important": "bold",
    }
)


class RichConsole:
    def __init__(self):
        self.console = rich.console.Console(
            highlight=False, theme=rich_theme, soft_wrap=True
        )
        self._status = None

    def info(self, *args, **kwargs):
        self.console.print(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.console.print(*args, **{"style": "warning", **kwargs})

    def error(self, *args, **kwargs):
        self.console.print(*args, **{"style": "error", **kwargs})

    def status(self, message, **kwargs_spinner_style):
        if not self._status:
            default = {}
            if self.is_advanced():
                default = {"spinner": "dots"}
            else:
                default = {"spinner": "line", "refresh_per_second": 6}
            st_args = {**default, **kwargs_spinner_style}

            self._status = self.console.status(message, **st_args)
            self._status.start()
        else:
            self._status.update(message, **kwargs_spinner_style)

    def stop_status(self):
        if self._status:
            self._status.stop()

    def log(self, *args, **kwargs):
        self.console.log(*args, **kwargs)

    def is_advanced(self):
        return not self.console.legacy_windows

    def info_table(self, rows, maxcolwidths):
        if not rows:
            return
        num_cols = len(rows[0])
        columns = [Column() for _ in range(num_cols)]

        if maxcolwidths:
            for i, c in enumerate(columns):
                c.width = maxcolwidths[i]

        table = Table.grid(*columns)
        for row in rows:
            table.add_row(*[str(c) for c in row])

        self.info(table)


class DebugConsole:
    def __init__(self):
        pass

    def info(self, message, *args, **kwargs):
        logger.info(message)

    def warning(self, message, *args, **kwargs):
        logger.warning(message)

    def error(self, message, *args, **kwargs):
        logger.error(message)

    def status(self, message, **kwargs_spinner_style):
        logger.warning(message)

    def stop_status(self):
        pass

    def log(self, message, *args, **kwargs):
        logger.info(message)

    def is_advanced(self):
        return False

    def info_table(self, rows, maxcolwidths):
        for row in rows:
            self.info("|".join([str(c) for c in row]))


class RootConsole:
    def __init__(self):
        # default
        self.console = DebugConsole()
        self.stack_status = deque()

    def info(self, message, *args, **kwargs):
        self.console.info(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.console.warning(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.console.error(message, *args, **kwargs)

    def status(self, message, clear=True, **kwargs_spinner_style):
        if clear:
            self.stack_status.clear()
        self.stack_status.append((message, kwargs_spinner_style))
        self.console.status(message, **kwargs_spinner_style)

    def pop_status(self):
        self.stack_status.pop()
        message, kwargs_spinner_style = self.stack_status[-1]
        self.console.status(message, **kwargs_spinner_style)

    def stop_status(self):
        self.console.stop_status()

    def log(self, message, *args, **kwargs):
        self.console.log(message, *args, **kwargs)

    def is_advanced(self):
        return self.console.is_advanced()

    def info_table(self, rows, maxcolwidths=None):
        self.console.info_table(rows, maxcolwidths)


ROOT_CONSOLE = RootConsole()


def getConsole():
    return ROOT_CONSOLE
